@@ -17,6 +17,11 @@ class RedisCache(BaseCache):
     _client_check_lock = Lock()
     _ttl = config.getint('cache', 'redis_ttl') or 60 * 60 * 12
 
+    def __init__(self, name, size_limit=1024, duration=_ttl, context=True):
+        super().__init__(
+            name, size_limit=size_limit, duration=duration, context=context)
+        self.ensure_client()
+
     @classmethod
     def ensure_client(cls):
         with cls._client_check_lock:
@@ -30,10 +35,6 @@ class RedisCache(BaseCache):
                 db = url.path.strip('/')
                 cls._client = redis.StrictRedis(host=host, port=port, db=db)
 
-    def __init__(self, name, size_limit=1024, duration=None, context=True):
-        super(RedisCache, self).__init__(name, size_limit, duration, context)
-        self.ensure_client()
-
     def _namespace(self, dbname=None):
         if dbname is None:
             dbname = Transaction().database.name
@@ -68,12 +69,17 @@ class RedisCache(BaseCache):
         for key in self._client.scan_iter(match='%s:*' % (namespace)):
             self._client.delete(key)
 
-    # XXX TODO: Transactional Cache
+    # TODO: Transactional Cache
     # https://discuss.tryton.org/t/transactional-cache/1012
     @classmethod
     def sync(cls, transaction):
         pass
 
+    # redis cache is synced immediately
+    # https://github.com/tryton/trytond/commit/9c7f4753221d8f5a8252b36d080e49bc54b9ef1d
+    def sync_since(self, value):
+        return False
+
     @classmethod
     def commit(cls, transaction):
         pass
