@@ -1,25 +1,24 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import time
 import datetime
-from xml import sax
 import logging
 import re
+import time
 
 from collections import defaultdict
 from decimal import Decimal
+from xml import sax
 
+from trytond import __version__
 from trytond.config import config
-
-from . import __version__
-from .tools import grouped_slice
-from .transaction import Transaction
-from .pyson import PYSONEncoder, CONTEXT
+from trytond.pyson import CONTEXT, PYSONEncoder
+from trytond.tools import grouped_slice
+from trytond.transaction import Transaction
 
 logger = logging.getLogger(__name__)
 
-CDATA_START = re.compile('^\s*\<\!\[cdata\[', re.IGNORECASE)
-CDATA_END = re.compile('\]\]\>\s*$', re.IGNORECASE)
+CDATA_START = re.compile(r'^\s*\<\!\[cdata\[', re.IGNORECASE)
+CDATA_END = re.compile(r'\]\]\>\s*$', re.IGNORECASE)
 
 _update_changed_fields = config.getboolean('optional',
     'update_changed_fields', default=True)
@@ -226,25 +225,26 @@ class RecordTagHandler:
             eval_attr = attributes.get('eval', '')
             pyson_attr = bool(int(attributes.get('pyson', '0')))
 
+            context = {}
+            context['time'] = time
+            context['version'] = __version__.rsplit('.', 1)[0]
+            context['ref'] = self.mh.get_id
+            context['Decimal'] = Decimal
+            context['datetime'] = datetime
+            if pyson_attr:
+                context.update(CONTEXT)
+
             if search_attr:
                 search_model = self.model._fields[field_name].model_name
                 SearchModel = self.mh.pool.get(search_model)
                 with Transaction().set_context(active_test=False):
-                    found, = SearchModel.search(eval(search_attr))
+                    found, = SearchModel.search(eval(search_attr, context))
                     self.values[field_name] = found.id
 
             elif ref_attr:
                 self.values[field_name] = self.mh.get_id(ref_attr)
 
             elif eval_attr:
-                context = {}
-                context['time'] = time
-                context['version'] = __version__.rsplit('.', 1)[0]
-                context['ref'] = self.mh.get_id
-                context['Decimal'] = Decimal
-                context['datetime'] = datetime
-                if pyson_attr:
-                    context.update(CONTEXT)
                 value = eval(eval_attr, context)
                 if pyson_attr:
                     value = PYSONEncoder(sort_keys=True).encode(value)
@@ -284,7 +284,7 @@ class RecordTagHandler:
                 self.cdata = 'done'
 
                 value = self.values[self.current_field]
-                match = re.findall('[^%]%\((.*?)\)[ds]', value)
+                match = re.findall(r'[^%]%\((.*?)\)[ds]', value)
                 xml_ids = {}
                 for xml_id in match:
                     xml_ids[xml_id] = self.mh.get_id(xml_id)
@@ -376,7 +376,7 @@ class Fs2bdAccessor:
                             cache, cache.get('_language_cache', {}).values()):
                         if (model_name in cache
                                 and model.id in cache[model_name]):
-                            cache[model_name][model.id] = {}
+                            del cache[model_name][model.id]
             self.browserecord[module][model_name][model.id] = model
 
     def fetch_new_module(self, module):
@@ -415,7 +415,7 @@ class Fs2bdAccessor:
 
 class TrytondXmlHandler(sax.handler.ContentHandler):
 
-    def __init__(self, pool, module, module_state, modules):
+    def __init__(self, pool, module, module_state, modules, languages):
         "Register known taghandlers, and managed tags."
         sax.handler.ContentHandler.__init__(self)
 
@@ -432,6 +432,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         self.grouped_model_data = []
         self.skip_data = False
         self.modules = modules
+        self.languages = languages
 
         # Tag handlders are used to delegate the processing
         self.taghandlerlist = {
@@ -459,11 +460,8 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
 
         try:
             self.sax_parser.parse(source)
-        except Exception:
-            logger.error(
-                "Error while parsing xml file:\n" + self.current_state(),
-                exc_info=True)
-            raise
+        except Exception as e:
+            raise Exception("Error " + self.current_state()) from e
         return self.to_delete
 
     def startElement(self, name, attributes):
@@ -477,15 +475,14 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
             elif name == "data":
                 self.noupdate = bool(int(attributes.get("noupdate", '0')))
                 self.grouped = bool(int(attributes.get('grouped', 0)))
-                if self.pool.test and \
-                        bool(int(attributes.get("skiptest", '0'))):
-                    self.skip_data = True
-                else:
-                    self.skip_data = False
+                self.skip_data = False
                 depends = attributes.get('depends', '').split(',')
                 depends = {m.strip() for m in depends if m}
                 if not depends.issubset(self.modules):
                     self.skip_data = True
+                if (attributes.get('language')
+                        and attributes.get('language') not in self.languages):
+                    self.skip_data = True
 
             elif name == "tryton":
                 pass
@@ -784,6 +781,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                             'db_id': record.id,
                             'values': self.ModelData.dump_values(values),
                             'fs_values': self.ModelData.dump_values(fs_values),
+                            'noupdate': self.noupdate,
                             }))
 
         # reset_browsercord to keep cache memory low
