@@ -1,7 +1,7 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import datetime
-from itertools import islice, chain, product, groupby
+from itertools import islice, chain, product, groupby, repeat
 from collections import OrderedDict, defaultdict
 from functools import wraps
 
@@ -164,6 +164,7 @@ class ModelSQL(ModelStorage):
     """
     Define a model with storage in database.
     """
+    __slots__ = ()
     _table = None  # The name of the table in database
     _order = None
     _order_name = None  # Use to force order field when sorting on Many2One
@@ -187,7 +188,7 @@ class ModelSQL(ModelStorage):
             cls._sql_constraints.append(
                 ('id_positive', Check(table, table.id >= 0),
                     'ir.msg_id_positive'))
-        cls._order = [('id', 'ASC')]
+        cls._order = [('id', None)]
         if issubclass(cls, ModelView):
             cls.__rpc__.update({
                     'history_revisions': RPC(),
@@ -208,13 +209,11 @@ class ModelSQL(ModelStorage):
 
     @classmethod
     def __table_handler__(cls, module_name=None, history=False):
-        TableHandler = backend.get('TableHandler')
-        return TableHandler(cls, module_name, history=history)
+        return backend.TableHandler(cls, module_name, history=history)
 
     @classmethod
     def __register__(cls, module_name):
         cursor = Transaction().connection.cursor()
-        TableHandler = backend.get('TableHandler')
         super(ModelSQL, cls).__register__(module_name)
 
         if callable(cls.table_query):
@@ -267,8 +266,8 @@ class ModelSQL(ModelStorage):
                             and not callable(ref_model.table_query)):
                         ref = ref_model._table
                         # Create foreign key table if missing
-                        if not TableHandler.table_exist(ref):
-                            TableHandler(ref_model)
+                        if not backend.TableHandler.table_exist(ref):
+                            backend.TableHandler(ref_model)
                     else:
                         ref = None
                 if field_name in ['create_uid', 'write_uid']:
@@ -337,7 +336,6 @@ class ModelSQL(ModelStorage):
     def __raise_integrity_error(
             cls, exception, values, field_names=None, transaction=None):
         pool = Pool()
-        TableHandler = backend.get('TableHandler')
         if field_names is None:
             field_names = list(cls._fields.keys())
         if transaction is None:
@@ -355,7 +353,7 @@ class ModelSQL(ModelStorage):
                         gettext('ir.msg_required_validation_record',
                             **cls.__names__(field_name)))
         for name, _, error in cls._sql_constraints:
-            if TableHandler.convert_name(name) in str(exception):
+            if backend.TableHandler.convert_name(name) in str(exception):
                 raise SQLConstraintError(gettext(error))
         # Check foreign key in last because this can raise false positive
         # if the target is created during the same transaction.
@@ -372,7 +370,8 @@ class ModelSQL(ModelStorage):
                 target_records = Model.search([
                         ('id', '=', field.sql_format(values[field_name])),
                         ], order=[])
-                if not ((target_records
+                if not ((
+                            target_records
                             or (values[field_name] in create_records))
                         and (values[field_name] not in delete_records)):
                     error_args = cls.__names__(field_name)
@@ -554,7 +553,6 @@ class ModelSQL(ModelStorage):
     @classmethod
     @no_table_query
     def create(cls, vlist):
-        DatabaseIntegrityError = backend.get('DatabaseIntegrityError')
         transaction = Transaction()
         cursor = transaction.connection.cursor()
         pool = Pool()
@@ -565,6 +563,7 @@ class ModelSQL(ModelStorage):
         table = cls.__table__()
         modified_fields = set()
         defaults_cache = {}  # Store already computed default values
+        missing_defaults = {}  # Store missing default values by schema
         new_ids = []
         vlist = [v.copy() for v in vlist]
         for values in vlist:
@@ -576,26 +575,29 @@ class ModelSQL(ModelStorage):
             modified_fields |= set(values.keys())
 
             # Get default values
-            default = []
-            for fname, field in cls._fields.items():
-                if fname in values:
-                    continue
-                if fname in [
-                        'create_uid', 'create_date',
-                        'write_uid', 'write_date', 'id']:
-                    continue
-                if isinstance(field, fields.Function) and not field.setter:
-                    continue
-                if fname in defaults_cache:
-                    values[fname] = defaults_cache[fname]
-                else:
-                    default.append(fname)
+            values_schema = tuple(sorted(values))
+            if values_schema not in missing_defaults:
+                default = []
+                missing_defaults[values_schema] = default_values = {}
+                for fname, field in cls._fields.items():
+                    if fname in values:
+                        continue
+                    if fname in [
+                            'create_uid', 'create_date',
+                            'write_uid', 'write_date', 'id']:
+                        continue
+                    if isinstance(field, fields.Function) and not field.setter:
+                        continue
+                    if fname in defaults_cache:
+                        default_values[fname] = defaults_cache[fname]
+                    else:
+                        default.append(fname)
 
-            if default:
-                defaults = cls.default_get(default, with_rec_name=False)
-                defaults = cls._clean_defaults(defaults)
-                values.update(defaults)
-                defaults_cache.update(defaults)
+                if default:
+                    defaults = cls.default_get(default, with_rec_name=False)
+                    default_values.update(cls._clean_defaults(defaults))
+                    defaults_cache.update(default_values)
+            values.update(missing_defaults[values_schema])
 
             insert_columns = [table.create_uid, table.create_date]
             insert_values = [transaction.user, CurrentTimestamp()]
@@ -625,7 +627,7 @@ class ModelSQL(ModelStorage):
                                 [insert_values]))
                         id_new = transaction.database.lastid(cursor)
                 new_ids.append(id_new)
-            except DatabaseIntegrityError as exception:
+            except backend.DatabaseIntegrityError as exception:
                 transaction = Transaction()
                 with Transaction().new_transaction(), \
                         Transaction().set_context(_check_access=False):
@@ -667,8 +669,9 @@ class ModelSQL(ModelStorage):
 
         cls._insert_history(new_ids)
 
-        field_names = list(cls._fields.keys())
-        cls._update_mptt(field_names, [new_ids] * len(field_names))
+        if cls._mptt_fields:
+            field_names = list(sorted(cls._mptt_fields))
+            cls._update_mptt(field_names, repeat(new_ids, len(field_names)))
 
         cls.__check_domain_rule(new_ids, 'create')
         records = cls.browse(new_ids)
@@ -695,8 +698,10 @@ class ModelSQL(ModelStorage):
 
         fields_related = defaultdict(set)
         extra_fields = set()
+        if 'write_date' not in fields_names:
+            extra_fields.add('write_date')
         for field_name in fields_names:
-            if field_name == '_timestamp':
+            if field_name in {'_timestamp', '_write', '_delete'}:
                 continue
             if '.' in field_name:
                 field_name, field_related = field_name.split('.', 1)
@@ -732,6 +737,28 @@ class ModelSQL(ModelStorage):
             field = cls._fields.get(f)
             if field and field.sql_type():
                 columns.append(field.sql_column(table).as_(f))
+                if backend.name == 'sqlite':
+                    columns[-1].output_name += ' [%s]' % field.sql_type().base
+            elif f in {'_write', '_delete'}:
+                if not callable(cls.table_query):
+                    rule_domain = Rule.domain_get(
+                        cls.__name__, mode=f.lstrip('_'))
+                    if rule_domain:
+                        rule_tables = {None: (table, None)}
+                        rule_tables, rule_expression = cls.search_domain(
+                            rule_domain, active_test=False, tables=rule_tables)
+                        if len(rule_tables) > 1:
+                            # The expression uses another table
+                            rule_tables, rule_expression = cls.search_domain(
+                                rule_domain, active_test=False)
+                            rule_from = convert_from(None, rule_tables)
+                            rule_table, _ = rule_tables[None]
+                            rule_where = rule_table.id == table.id
+                            rule_expression = rule_from.select(
+                                        rule_expression, where=rule_where)
+                        columns.append(rule_expression.as_(f))
+                    else:
+                        columns.append(Literal(True).as_(f))
             elif f == '_timestamp' and not callable(cls.table_query):
                 sql_type = fields.Char('timestamp').sql_type().base
                 columns.append(Extract('EPOCH',
@@ -768,17 +795,21 @@ class ModelSQL(ModelStorage):
             result = [{'id': x} for x in ids]
 
         cachable_fields = []
+        max_write_date = max(
+            (r['write_date'] for r in result if r.get('write_date')),
+            default=None)
         for column in columns:
             # Split the output name to remove SQLite type detection
             fname = column.output_name.split()[0]
-            if fname == '_timestamp':
+            if fname.startswith('_'):
                 continue
             field = cls._fields[fname]
             if not hasattr(field, 'get'):
                 if getattr(field, 'translate', False):
                     translations = Translation.get_ids(
                         cls.__name__ + ',' + fname, 'model',
-                        Transaction().language, ids)
+                        Transaction().language, ids,
+                        cached_after=max_write_date)
                     for row in result:
                         row[fname] = translations.get(row['id']) or row[fname]
                 if fname != 'id':
@@ -790,10 +821,8 @@ class ModelSQL(ModelStorage):
 
         if getter_fields and cachable_fields:
             cache = transaction.get_cache().setdefault(
-                cls.__name__, LRUDict(cache_size()))
+                cls.__name__, LRUDict(cache_size(), cls._record))
             for row in result:
-                if row['id'] not in cache:
-                    cache[row['id']] = {}
                 for fname in cachable_fields:
                     cache[row['id']][fname] = row[fname]
 
@@ -832,11 +861,15 @@ class ModelSQL(ModelStorage):
                         date_result = date_results[fname]
                         row[fname] = date_result[row['id']]
             else:
-                getter_results = field.get(ids, cls, field_list, values=result)
-                for fname in field_list:
-                    getter_result = getter_results[fname]
-                    for row in result:
-                        row[fname] = getter_result[row['id']]
+                for sub_results in grouped_slice(result, cache_size()):
+                    sub_results = list(sub_results)
+                    sub_ids = [r['id'] for r in sub_results]
+                    getter_results = field.get(
+                        sub_ids, cls, field_list, values=sub_results)
+                    for fname in field_list:
+                        getter_result = getter_results[fname]
+                        for row in sub_results:
+                            row[fname] = getter_result[row['id']]
 
         def read_related(field, Target, rows, fields):
             name = field.name
@@ -928,7 +961,6 @@ class ModelSQL(ModelStorage):
     @classmethod
     @no_table_query
     def write(cls, records, values, *args):
-        DatabaseIntegrityError = backend.get('DatabaseIntegrityError')
         transaction = Transaction()
         cursor = transaction.connection.cursor()
         pool = Pool()
@@ -981,7 +1013,7 @@ class ModelSQL(ModelStorage):
                 try:
                     cursor.execute(*table.update(columns, update_values,
                             where=red_sql))
-                except DatabaseIntegrityError as exception:
+                except backend.DatabaseIntegrityError as exception:
                     transaction = Transaction()
                     with Transaction().new_transaction(), \
                             Transaction().set_context(_check_access=False):
@@ -1000,9 +1032,12 @@ class ModelSQL(ModelStorage):
                 if hasattr(field, 'set'):
                     fields_to_set.setdefault(fname, []).extend((ids, value))
 
-            field_names = list(values.keys())
-            cls._update_mptt(field_names, [ids] * len(field_names), values)
-            all_field_names |= set(field_names)
+            mptt_fields = cls._mptt_fields & set(values)
+            if mptt_fields:
+                cls._update_mptt(
+                    list(sorted(mptt_fields)), repeat(ids, len(mptt_fields)),
+                    values)
+            all_field_names |= set(values)
 
         for fname in sorted(fields_to_set, key=cls.index_set_field):
             fargs = fields_to_set[fname]
@@ -1020,7 +1055,6 @@ class ModelSQL(ModelStorage):
     @classmethod
     @no_table_query
     def delete(cls, records):
-        DatabaseIntegrityError = backend.get('DatabaseIntegrityError')
         transaction = Transaction()
         cursor = transaction.connection.cursor()
         pool = Pool()
@@ -1041,20 +1075,18 @@ class ModelSQL(ModelStorage):
         cls.__check_timestamp(ids)
         cls.__check_domain_rule(ids, 'delete')
 
-        has_translation = False
         tree_ids = {}
-        for fname, field in cls._fields.items():
-            if (isinstance(field, fields.Many2One)
-                    and field.model_name == cls.__name__
-                    and field.left and field.right):
-                tree_ids[fname] = []
-                for sub_ids in grouped_slice(ids):
-                    where = reduce_ids(field.sql_column(table), sub_ids)
-                    cursor.execute(*table.select(table.id, where=where))
-                    tree_ids[fname] += [x[0] for x in cursor.fetchall()]
-            if (getattr(field, 'translate', False)
-                    and not hasattr(field, 'set')):
-                has_translation = True
+        for fname in cls._mptt_fields:
+            field = cls._fields[fname]
+            tree_ids[fname] = []
+            for sub_ids in grouped_slice(ids):
+                where = reduce_ids(field.sql_column(table), sub_ids)
+                cursor.execute(*table.select(table.id, where=where))
+                tree_ids[fname] += [x[0] for x in cursor]
+
+        has_translation = any(
+            getattr(f, 'translate', False) and not hasattr(f, 'set')
+            for f in cls._fields.values())
 
         foreign_keys_tocheck = []
         foreign_keys_toupdate = []
@@ -1087,7 +1119,7 @@ class ModelSQL(ModelStorage):
                     Column(foreign_table, field_name), sub_ids)
                 cursor.execute(*foreign_table.select(foreign_table.id,
                         where=foreign_red_sql))
-                records = Model.browse([x[0] for x in cursor.fetchall()])
+                records = Model.browse([x[0] for x in cursor])
             else:
                 with transaction.set_context(active_test=False):
                     records = Model.search([(field_name, 'in', sub_ids)])
@@ -1134,7 +1166,7 @@ class ModelSQL(ModelStorage):
 
             try:
                 cursor.execute(*table.delete(where=red_sql))
-            except DatabaseIntegrityError as exception:
+            except backend.DatabaseIntegrityError as exception:
                 transaction = Transaction()
                 with Transaction().new_transaction():
                     cls.__raise_integrity_error(
@@ -1203,12 +1235,7 @@ class ModelSQL(ModelStorage):
         if wrong_ids:
             model = cls.__name__
             if Model:
-                models = Model.search([
-                        ('model', '=', cls.__name__),
-                        ], limit=1)
-                if models:
-                    model, = models
-                    model = model.name
+                model = Model.get_name(cls.__name__)
             ids = ', '.join(map(str, ids[:5]))
             if len(wrong_ids) > 5:
                 ids += '...'
@@ -1262,11 +1289,14 @@ class ModelSQL(ModelStorage):
         for oexpr, otype in order:
             fname, _, extra_expr = oexpr.partition('.')
             field = cls._fields[fname]
-            otype = otype.upper()
-            try:
-                otype, null_ordering = otype.split(' ', 1)
-            except ValueError:
-                null_ordering = None
+            if not otype:
+                otype, null_ordering = 'ASC', None
+            else:
+                otype = otype.upper()
+                try:
+                    otype, null_ordering = otype.split(' ', 1)
+                except ValueError:
+                    null_ordering = None
             Order = order_types[otype]
             NullOrdering = null_ordering_types[null_ordering]
             forder = field.convert_order(oexpr, tables, cls)
@@ -1315,7 +1345,7 @@ class ModelSQL(ModelStorage):
         rows = list(cursor_dict(cursor, transaction.database.IN_MAX))
         cache = transaction.get_cache()
         if cls.__name__ not in cache:
-            cache[cls.__name__] = LRUDict(cache_size())
+            cache[cls.__name__] = LRUDict(cache_size(), cls._record)
         delete_records = transaction.delete_records.setdefault(cls.__name__,
             set())
 
@@ -1346,7 +1376,7 @@ class ModelSQL(ModelStorage):
                         & (history.create_date == Null)
                         & (history.write_date
                             <= transaction.context['_datetime'])))
-                for deleted_id, delete_date in cursor.fetchall():
+                for deleted_id, delete_date in cursor:
                     history_date, _ = ids_history[deleted_id]
                     if isinstance(history_date, str):
                         strptime = datetime.datetime.strptime
@@ -1371,7 +1401,6 @@ class ModelSQL(ModelStorage):
                     keys = list(data.keys())
                     for k in keys[:]:
                         if k in ('_timestamp', '_datetime', '__id'):
-                            keys.remove(k)
                             continue
                         field = cls._fields[k]
                         if not getattr(field, 'datetime_field', None):
@@ -1379,7 +1408,7 @@ class ModelSQL(ModelStorage):
                             continue
                 for k in keys:
                     del data[k]
-                cache[cls.__name__].setdefault(data['id'], {}).update(data)
+                cache[cls.__name__][data['id']]._update(data)
 
         if len(rows) >= transaction.database.IN_MAX:
             if (cls._history
@@ -1428,7 +1457,8 @@ class ModelSQL(ModelStorage):
                 return And((convert(d) for d in (
                             domain[1:] if domain[0] == 'AND' else domain)))
 
-        expression = convert(domain)
+        with Transaction().set_context(_check_access=False):
+            expression = convert(domain)
 
         if cls._history and transaction.context.get('_datetime'):
             table, _ = tables[None]
@@ -1445,35 +1475,31 @@ class ModelSQL(ModelStorage):
         cursor = transaction.connection.cursor()
         for field_name, ids in zip(field_names, list_ids):
             field = cls._fields[field_name]
-            if (isinstance(field, fields.Many2One)
-                    and field.model_name == cls.__name__
-                    and field.left and field.right):
-                if (values is not None
-                        and (field.left in values or field.right in values)):
-                    raise Exception('ValidateError',
-                        'You can not update fields: "%s", "%s"' %
-                        (field.left, field.right))
+            if (values is not None
+                    and (field.left in values or field.right in values)):
+                raise Exception('ValidateError',
+                    'You can not update fields: "%s", "%s"' %
+                    (field.left, field.right))
 
-                # Nested creation require a rebuild
-                # because initial values are 0
-                # and thus _update_tree can not find the children
-                table = cls.__table__()
-                parent = cls.__table__()
-                cursor.execute(*table.join(parent,
-                        condition=Column(table, field_name) == parent.id
-                        ).select(table.id,
-                        where=(Column(parent, field.left) == 0)
-                        & (Column(parent, field.right) == 0),
-                        limit=1))
-                nested_create = cursor.fetchone()
+            # Nested creation require a rebuild
+            # because initial values are 0
+            # and thus _update_tree can not find the children
+            table = cls.__table__()
+            parent = cls.__table__()
+            cursor.execute(*table.join(parent,
+                    condition=Column(table, field_name) == parent.id
+                    ).select(table.id,
+                    where=(Column(parent, field.left) == 0)
+                    & (Column(parent, field.right) == 0),
+                    limit=1))
+            nested_create = cursor.fetchone()
 
-                if (not nested_create and len(ids) < 2 and not
-                        context.get('force_update_mptt')):
-                    for id_ in ids:
-                        cls._update_tree(id_, field_name,
-                            field.left, field.right)
-                else:
-                    cls._rebuild_tree(field_name, None, 0)
+            if not nested_create and len(ids) < 2:
+                for id_ in ids:
+                    cls._update_tree(id_, field_name,
+                        field.left, field.right)
+            else:
+                cls._rebuild_tree(field_name, None, 0)
 
     @classmethod
     def _rebuild_tree(cls, parent, parent_id, left):
@@ -1486,9 +1512,7 @@ class ModelSQL(ModelStorage):
 
         cursor.execute(*table.select(table.id,
                 where=Column(table, parent) == parent_id))
-        childs = cursor.fetchall()
-
-        for child_id, in childs:
+        for child_id, in cursor:
             right = cls._rebuild_tree(parent, child_id, right)
 
         field = cls._fields[parent]
@@ -1587,7 +1611,7 @@ class ModelSQL(ModelStorage):
                     cursor.execute(*table.select(*columns, where=where))
 
                     where = Literal(False)
-                    for row in cursor.fetchall():
+                    for row in cursor:
                         clause = table.id != row[0]
                         for column, operator, value in zip(
                                 sql.columns, sql.operators, row[1:]):
