@@ -78,6 +78,11 @@ def login(dbname, loginname, parameters, cache=True, context=None):
 
 
 def logout(dbname, user, session, context=None):
+    if config_session_redis():
+        name = redis.get_session(dbname, user, session)
+        if name:
+            redis.del_session(dbname, user, session)
+        return name
     for count in range(config.getint('database', 'retry'), -1, -1):
         with Transaction().start(dbname, 0, context=context):
             pool = _get_pool(dbname)
@@ -98,6 +103,13 @@ def logout(dbname, user, session, context=None):
 
 
 def check(dbname, user, session, context=None):
+    if config_session_redis():
+        ttl = redis.hit_session(dbname, user, session)
+        if ttl is not None:
+            if config_session_audit():
+                redis.time_user(dbname, user, ttl)
+            return user
+        return
     for count in range(config.getint('database', 'retry'), -1, -1):
         with Transaction().start(dbname, user, context=context) as transaction:
             pool = _get_pool(dbname)
